package refit.replica;

import refit.application.REFITApplicationServer;
import refit.application.REFITSystemApplicationServer;
import refit.config.REFITConfig;
import refit.message.REFITMessageType;
import refit.replica.checkpoint.REFITCheckpointStage;
import refit.replica.client.REFITClientStage;
import refit.replica.coupling.REFITCouplingHelper;
import refit.replica.coupling.REFITCouplingStage;
import refit.replica.execution.REFITExecutionStage;
import refit.replica.execution.REFITExecutionStageBase;
import refit.replica.execution.REFITExecutionStubStage;
import refit.replica.fuzzycheckpoint.REFITFuzzyCheckpointStage;
import refit.replica.order.REFITOrderGroups;
import refit.replica.order.REFITOrderStage;
import refit.replica.update.REFITUpdateStage;
import refit.scheduler.REFITScheduler;
import refit.scheduler.REFITSchedulerTaskType;

public class REFITOmadaReplica extends REFITBaseReplica {
	private final REFITClientStage[] clientStage;
	private final REFITOrderStage[] orderStage;
	private final REFITCouplingStage[] couplingStage;

	private final REFITUpdateStage updateStage;
	private final REFITCheckpointStage checkpointStage;
	private final REFITExecutionStageBase executionStage;
	private final REFITFuzzyCheckpointStage fuzzyCheckpointStage;

	private final REFITGroupMessageRouter msgRouter;


	public REFITOmadaReplica(short id, boolean isSingleThreaded) {
		super(id, isSingleThreaded, true);

		msgRouter = new REFITGroupMessageRouter(this);
		addToClientInbound(msgRouter);
		addToInbound(REFITMessageType.ORDER, msgRouter);
		addToInbound(REFITMessageType.ORDERED_REQUEST, msgRouter);

		// Create global stages
		if (id < REFITConfig.EXEC_GROUP_SIZE) {
			// create and initialize application
			final REFITApplicationServer application = createApplication();
			application.init();
			final REFITSystemApplicationServer systemApplication = createSystemApplication();
			systemApplication.init();

			this.executionStage = new REFITExecutionStage(this, application, systemApplication);
			this.fuzzyCheckpointStage = (REFITConfig.ENABLE_DIFFERENTIAL_CHECKPOINTING || REFITConfig.ENABLE_FUZZY_CHECKPOINTS)
					? new REFITFuzzyCheckpointStage(this, application) : null;
		} else {
			this.executionStage = new REFITExecutionStubStage(this);
			this.fuzzyCheckpointStage = null;
		}
		this.updateStage = REFITConfig.USE_PASSIVE_REPLICATION ? new REFITUpdateStage(this, REFITReplicaGroups.getExecutionReplicas(0).clone()) : null;
		if (updateStage != null) {
			addToInbound(REFITMessageType.UPDATE, updateStage);
			addToInbound(REFITMessageType.BATCH_UPDATE, updateStage);
		}
		this.checkpointStage = new REFITCheckpointStage(this, REFITReplicaGroups.getExecutionReplicas(0),
				REFITReplicaGroups.getAllReplicas(), null);
		addToInbound(REFITMessageType.CHECKPOINT, checkpointStage);
		addToInbound(REFITMessageType.CHECKPOINT_STATE, checkpointStage);

		// Create group stages
		clientStage = new REFITClientStage[REFITOrderGroups.COUNT];
		orderStage = new REFITOrderStage[REFITOrderGroups.COUNT];
		couplingStage = new REFITCouplingStage[REFITOrderGroups.COUNT];
		REFITCouplingHelper helper = new REFITCouplingHelper(this);
		addToInbound(REFITMessageType.EXECUTOR_FLUSH, helper);

		for (short groupID = 0; groupID < REFITOrderGroups.COUNT; groupID++) {
			boolean inGroup = REFITOrderGroups.isInGroupByID(id, groupID);
			if (!inGroup && id >= REFITConfig.EXEC_GROUP_SIZE) {
				continue;
			}

			// also setup client and coupling stage on execution replicas
			clientStage[groupID] = new REFITClientStage(this, groupID);
			msgRouter.requestSenders[groupID].addListener(clientStage[groupID]);
			executionStage.replySenders[groupID].addListener(clientStage[groupID]);

			couplingStage[groupID] = new REFITCouplingStage(this, helper, groupID);
			msgRouter.executeSenders[groupID].addListener(couplingStage[groupID]);
			couplingStage[groupID].requestSender.addListener(executionStage);
			if (id < REFITConfig.EXEC_GROUP_SIZE) {
				clientStage[groupID].readRequestSender.addListener(executionStage);
			}
			if (!inGroup) {
				continue;
			}

			orderStage[groupID] = new REFITOrderStage(this, REFITConfig.INITIAL_ORDER_PROTOCOL, groupID, null);
			msgRouter.orderSenders[groupID].addListener(orderStage[groupID]);
			helper.flushRequestSender.addListener(orderStage[groupID]);
			helper.awaitRequestSender.addListener(clientStage[groupID]);

			clientStage[groupID].requestSender.addListener(orderStage[groupID]);
			clientStage[groupID].panicSender.addListener(orderStage[groupID]);
			// Inform about partially ordered requests
			couplingStage[groupID].requestSender.addListener(clientStage[groupID]);

			orderStage[groupID].requestSender.addListener(couplingStage[groupID]);
			orderStage[groupID].viewChangeSender.addListener(clientStage[groupID]);
			orderStage[groupID].viewChangeSender.addListener(couplingStage[groupID]);
			orderStage[groupID].viewChangeSender.addListener(executionStage);
			if (updateStage != null) {
				orderStage[groupID].viewChangeSender.addListener(updateStage);
			}

			checkpointStage.progressSender.addListener(orderStage[groupID]);
		}

		executionStage.checkpointSender.addListener(checkpointStage);
		checkpointStage.progressSender.addListener(executionStage);
		checkpointStage.stableCPSender.addListener(executionStage);
		checkpointStage.stableCPStateSender.addListener(executionStage);

		if (fuzzyCheckpointStage != null) {
			executionStage.updateSender.addListener(fuzzyCheckpointStage);
			executionStage.fuzzyCheckpointNotificationSender.addListener(fuzzyCheckpointStage);

			fuzzyCheckpointStage.fuzzyCheckpointNotificationSender.addListener(executionStage);
			fuzzyCheckpointStage.checkpointSender.addListener(checkpointStage);
		}

		if (updateStage != null) {
			checkpointStage.progressSender.addListener(updateStage);
			executionStage.updateSender.addListener(updateStage);
			updateStage.stableUpdateSender.addListener(executionStage);
		}

		// treat message as directly received from a client
		addClientInbound(REFITMessageType.REQUEST);
	}

	@Override
	protected void assignTasks(REFITScheduler[] schedulers) {
		// Assign tasks to schedulers
		schedulers[1 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.NETWORK_ENDPOINT);
		schedulers[1 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.CLIENT_NETWORK_ENDPOINT_WORKER);
		schedulers[1 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.CLIENT_STAGE);
		schedulers[0 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.REPLICA_NETWORK_ENDPOINT_WORKER);
		schedulers[0 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.ORDER_STAGE);
		schedulers[2 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.CHECKPOINT_STAGE);
		schedulers[3 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.COUPLING_STAGE);
		schedulers[3 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.EXECUTION_STAGE);
		schedulers[3 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.UPDATE_STAGE);
		schedulers[4 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.FUZZY_CHECKPOINT_STAGE);
		schedulers[0 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.PROBE_STAGE);
	}
}
