package refit.replica.client;

import java.util.ArrayList;

import refit.client.REFITClientLibrary;
import refit.config.REFITConfig;
import refit.config.REFITLogger;
import refit.message.*;
import refit.message.REFITInstruction.*;
import refit.replica.REFITBaseReplica;
import refit.replica.REFITObserver;
import refit.replica.REFITStage;
import refit.replica.order.REFITOrderGroups;
import refit.replica.order.REFITOrderProtocol;
import refit.scheduler.REFITSchedulerTaskType;
import refit.util.REFITTime;


public class REFITClientStage extends REFITStage {

	public final REFITObserver<REFITRequest> requestSender;
	public final REFITObserver<REFITRequest> readRequestSender;
	public final REFITObserver<REFITPanicNotification> panicSender;
	public final REFITObserver<REFITRequest> probeSender;

	private final boolean wrappedRequests;

	private final REFITReplyCache replyCache;
	private int groupID;
	private boolean sendReplies;
	private short contactReplica;
	private long executionNextSeqNr = 0;

	private long latestPanicNotification;
	private int panicInterval;
	private int requestPanicTimeout;


	public REFITClientStage(REFITBaseReplica replica, int groupID) {
		super(REFITSchedulerTaskType.CLIENT_STAGE, replica);
		markInterleaved();
		this.replyCache = new REFITReplyCache();
		this.groupID = groupID;
		this.sendReplies = true;
		this.contactReplica = -1;
		wrappedRequests = REFITConfig.SIGNED_REQUESTS && groupID != -1;

		requestSender = new REFITObserver<>();
		readRequestSender = new REFITObserver<>();
		probeSender = new REFITObserver<>();
		panicSender = new REFITObserver<>();

		resetTimeouts();
	}

	@Override
	public String toString() {
		return "CLINT[" + groupID + "]";
	}

	@Override
	protected void handleMessage(REFITMessage message) {
		switch (message.type) {
		case REQUEST:
			if (!wrappedRequests) {
				handleRequest((REFITRequest) message);
			} else {
				REFITLogger.logError(this, "drop unexpected request");
			}
			break;
		case REQUEST_WRAPPER:
			if (wrappedRequests) {
				handleRequestWrapper((REFITRequestWrapper) message);
			} else {
				REFITLogger.logError(this, "drop unexpected request wrapper");
			}
			break;
		case REPLY:
		case ACK_REPLY:
			handleReply((REFITReplyBase) message);
			break;
		case INSTRUCTION:
			handleInstruction((REFITInstruction) message);
			break;
		default:
			REFITLogger.logError(this, "drop message of unexpected type " + message.type);
		}
	}

	private void handleRequestWrapper(REFITRequestWrapper wrapper) {
		if (!replica.mac.verifyUnicastMAC(wrapper, REFITMessageAuthentication.ANY_SENDER)) return;

		handleRequest(wrapper.getRequest());
	}

	private void handleRequest(REFITRequest request) {
		if (REFITConfig.ENABLE_DEBUG_CHECKS && groupID >= 0 && request.groupID != groupID) {
			REFITLogger.logError(this, "Request for wrong group " + request + ", " + request.groupID + " expected " + groupID);
			return;
		}

		// Ignore old requests
		if (replyCache.isOldUID(request.uid)) return;

		// Resend reply if already available
		// The execution replicas are not always able to verify the request
		REFITReplyBase reply = replyCache.getReply(request.uid);
		if (reply != null) {
			sendReply(reply, true);
			return;
		}

		// Autonomous panics: Ignore request if it was already processed
		if (REFITConfig.AUTONOMOUS_PANICS && replyCache.isCurrentUID(request.uid)) return;

		// Don't handle faulty requests
		if (wrappedRequests && !request.mode.isTotalOrder()) {
			replica.mac.verifyNullMAC(request);
		} else {
			if (REFITConfig.SIGNED_REQUESTS && !replica.mac.verifySignature(request, REFITMessageAuthentication.ANY_SENDER)
					|| !REFITConfig.SIGNED_REQUESTS && !replica.mac.verifyMulticastMAC(request, REFITMessageAuthentication.ANY_SENDER))
				return;
		}

		// Ignore duplicate non-panic requests
		boolean isNewUID = replyCache.registerRequest(request.uid, request);
		if (isNewUID) {
			forwardRequest(request);
		} else if (request.isPanic) {
			// only have one group process the requests of each client
			if (!REFITConfig.LOCATION_BASED_ORDER_GROUPS && !REFITOrderGroups.isInGroup(replica.id, request.from))
				return;

			// read-only requests must be resent as write requests, so ignore them
			if (request.mode.isReadOnly()) return;

			// ignore panic for requests that cannot be executed just yet, the client will resend the panic anyways
			if (request.executeAfter >= executionNextSeqNr) {
				REFITLogger.logWarning(this, "Ignoring panic with executeAfter "
						+ request.executeAfter + " newer than current progress " + executionNextSeqNr);
				return;
			}

			if (replyCache.isUIDTimeOut(request.uid, requestPanicTimeout)) {
				triggerPanic();
			}
		}
	}

	private void forwardRequest(REFITRequest request) {
		if (request.mode.equals(REFITClientLibrary.RequestMode.PROBE)) {
			probeSender.broadcast(request);
		} else if (request.mode.isTotalOrder()
				&& !(REFITConfig.USE_PBFT_READ_OPTIMIZATION && request.mode.equals(REFITClientLibrary.RequestMode.READ))) {
			if (replica.id != contactReplica) {
				if (wrappedRequests) {
					REFITRequestWrapper wrapper = new REFITRequestWrapper(request.uid, replica.id, request);
					wrapper.serializeMessage();
					replica.mac.appendUnicastMAC(wrapper, contactReplica);

					replica.replicaUnicast(wrapper, contactReplica);
				} else {
					replica.replicaUnicast(request, contactReplica);
				}
			} else {
				// Forward request to local order stage
				requestSender.broadcast(request);
			}
		} else {
			// forward read to execution
			readRequestSender.broadcast(request);
		}
	}

	private void handleReply(REFITReplyBase reply) {
		if (REFITConfig.COLLECT_MESSAGE_STATISTICS) reply.getContext().statistics.showTrace();

		// prevent overwriting published replies
		REFITReplyBase oldReply = replyCache.getReply(reply.uid);
		if (oldReply != null && oldReply.type == REFITMessageType.REPLY && !((REFITReply) oldReply).isInternalOnly) {
			if (reply.type == REFITMessageType.REPLY && !((REFITReply) reply).isInternalOnly)
				REFITLogger.logWarning(this, "Trying to send a second reply to a client " + reply + " already sent " + oldReply);
			return;
		}

		replyCache.storeReply(reply.uid, reply);

		// without authentication (=CFT mode) the client just needs a single reply
		if (sendReplies && (REFITConfig.AUTHENTICATE_MESSAGES || -1 == reply.replyReplica
				|| replica.id == reply.replyReplica)) {
			sendReply(reply, false);
		}
	}

	private void sendReply(REFITReplyBase reply, boolean retry) {
		if (reply.type == REFITMessageType.REPLY) {
			REFITReply rep = (REFITReply) reply;
			if (rep.isInternalOnly) return;
			if (REFITConfig.USE_HASHED_REPLIES && rep.sendHashReply) {
				reply = rep.createHashedReply();
			}
		}
		if (retry)
			REFITLogger.logWarning(this, "resend reply " + reply);
		boolean serialized = reply.serializeMessage();
		if (serialized) replica.mac.appendUnicastMAC(reply, reply.uid.nodeID);
		replica.sendMessageToClient(reply, reply.uid.nodeID);
	}

	private void handleInstruction(REFITInstruction instruction) {
		switch (instruction.instructionType) {
		case CONFIGURATION_NOTIFICATION:
			REFITConfigurationNotification notification = (REFITConfigurationNotification) instruction;
			int limitedGroupID = Math.max(0, groupID);
			if (limitedGroupID != notification.groupID) {
				throw new InternalError("Got configuration for unexpected group " + notification.groupID);
			}

			REFITOrderProtocol protocol = notification.protocol;
			sendReplies = protocol.getActiveOrderRecipients(limitedGroupID)[replica.id];
			contactReplica = REFITOrderGroups.getGroupPrimary(notification.groupID, notification.viewID);

			if (REFITConfig.SIGNED_REQUESTS) {
				// Flush incomplete requests to the order stage / new contact replica, only works with signatures
				ArrayList<REFITRequest> requests = replyCache.flushIncomplete();
				for (REFITRequest request : requests) {
					forwardRequest(request);
				}
			} else {
				// when using mac authenticators, we need a way to get rid of requests with <= f valid macs which would block the agreement.
				// Allow clients to submit incomplete requests a second time after a panic
				replyCache.clearIncomplete();
			}
			resetTimeouts();
			REFITLogger.logEvent(this, "Configuration update (send replies: " + sendReplies + " contactReplica: " + contactReplica + ")");
			break;
		case INTERNAL_VIEW_CHANGE_NOTIFICATION:
			// reset the view change timers, but keep the timeout as this is a cooperative view change
			latestPanicNotification = System.currentTimeMillis();
			break;
		case EXECUTION_PROGRESS_NOTIFICATION:
			REFITExecutionProgressNotification lastSeq = (REFITExecutionProgressNotification) instruction;
			executionNextSeqNr = lastSeq.nextSeqNr;
			break;
		case AWAIT_REQUEST:
			handleAwaitRequest((REFITAwaitRequestNotification) instruction);
			if (!hasTimeout()) {
				setTimeout(50);
			}
			break;
		case EXECUTE_REQUEST:
			handleExecuteRequest(((REFITExecuteRequest) instruction));
			break;
		default:
			REFITLogger.logError(this, "drop instruction of unexpected type " + instruction.instructionType);
		}
	}

	@Override
	public void init() {
		super.init();
		if (REFITConfig.AUTONOMOUS_PANICS) {
			setTimeout(50);
		}
	}

	@Override
	public void timeout() {
		super.timeout();
		// just check whether a panic should be triggered in regular intervals
		// This avoids all the complications related to rescheduling the check
		setTimeout(50);

		// don't panic if replica itself is the leader
		if (contactReplica == replica.id || !replyCache.shouldPanic(REFITConfig.REPLICA_REQUEST_TIMEOUT)) return;

		triggerPanic();
	}

	private void triggerPanic() {
		// Control the rate at which panic notifications are sent
		if (REFITTime.currentTimeMillis.getAsLong() < latestPanicNotification + panicInterval) return;

		// Forward panic to order stage
		panicSender.broadcast(new REFITPanicNotification());
		// Update panic information
		latestPanicNotification = REFITTime.currentTimeMillis.getAsLong();
		panicInterval *= 2;
		requestPanicTimeout *= 2;
	}

	private void resetTimeouts() {
		latestPanicNotification = REFITTime.currentTimeMillis.getAsLong();
		panicInterval = REFITConfig.REPLICA_PANIC_INTERVAL;
		requestPanicTimeout = REFITConfig.REPLICA_REQUEST_TIMEOUT;
	}

	// Reuse client panic handling to enforce  that order groups make progress
	private void handleAwaitRequest(REFITAwaitRequestNotification await) {
		// fake but nevertheless unique ID
		REFITUniqueID fakeUid = new REFITUniqueID(replica.id, await.agreementSeqNr);
		// register to cause a panic if no reply is generated
		replyCache.registerRequest(fakeUid, null);
	}

	private void handleExecuteRequest(REFITExecuteRequest execute) {
		// treat seq# of other groups as reached
		long reachedSeqNr = REFITOrderGroups.nextSequenceNumber(execute.agreementSeqNr, groupID) - 1;
		REFITUniqueID fakeUid = new REFITUniqueID(replica.id, reachedSeqNr);
		// insert fake reply to stop panic timers
		replyCache.storeReply(fakeUid, null);
	}
}
