package refit.replica.coupling;

import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;

import refit.config.REFITConfig;
import refit.config.REFITLogger;
import refit.message.REFITInstruction;
import refit.message.REFITInstruction.REFITConfigurationNotification;
import refit.message.REFITInstruction.REFITExecuteRequest;
import refit.message.REFITMessage;
import refit.replica.REFITBaseReplica;
import refit.replica.REFITObserver;
import refit.replica.REFITReplicaGroups;
import refit.replica.REFITStage;
import refit.replica.checkpoint.REFITCheckpoint;
import refit.replica.order.REFITOrderGroups;
import refit.scheduler.REFITSchedulerTaskType;

public class REFITCouplingStage extends REFITStage {

	public final REFITObserver<REFITExecuteRequest> requestSender;

	private REFITCouplingHelper helper;

	private final boolean[] executionReplicas;
	private final boolean[] nonGroupExecutionReplicas;
	private final short nonGroupReplicaCount;
	private int groupID;
	private short contactReplica;

	public REFITCouplingStage(REFITBaseReplica replica, REFITCouplingHelper helper, int groupID) {
		super(REFITSchedulerTaskType.COUPLING_STAGE, replica);
		requestSender = new REFITObserver<>();

		this.helper = helper;
		this.groupID = groupID;
		orderCertificates = new HashMap<>();
		nextAgreementSeqNr = REFITOrderGroups.nextSequenceNumber(-1, groupID);
		highestReceivedSeqNr = -1;
		executionReplicas = REFITReplicaGroups.getExecutionReplicas(0);

		if (REFITConfig.OPTIMIZED_AGREEMENT_AND_EXECUTION) {
			boolean[] replicas = Arrays.copyOf(executionReplicas, executionReplicas.length);
			boolean[] groupReplicas = REFITOrderGroups.getGroupByID(groupID);
			for (int i = 0; i < groupReplicas.length; i++) {
				if (groupReplicas[i]) {
					replicas[i] = false;
				}
			}
			short replicaCount = 0;
			for (boolean isExecutionReplica : replicas) {
				replicaCount += (isExecutionReplica) ? 1 : 0;
			}
			nonGroupExecutionReplicas = (replicaCount > 0) ? replicas : null;
			nonGroupReplicaCount = replicaCount;
		} else {
			nonGroupExecutionReplicas = null;
			nonGroupReplicaCount = 0;
		}

		agreementProgress = false;
	}

	@Override
	public String toString() {
		return "COUPL[" + groupID + "]";
	}

	@Override
	protected void handleMessage(REFITMessage message) {
		switch (message.type) {
		case INSTRUCTION:
			handleInstruction((REFITInstruction) message);
			break;
		case ORDERED_REQUEST:
			handleOrderedRequest((REFITOrderedRequest) message);
			break;
		case CHECKPOINT:
			handleCheckpoint((REFITCheckpoint) message);
			break;
		default:
			REFITLogger.logError(this, "drop message of unexpected type " + message.type);
		}
	}

	private void handleInstruction(REFITInstruction instruction) {
		switch (instruction.instructionType) {
		case CONFIGURATION_NOTIFICATION:
			REFITConfigurationNotification notification = (REFITConfigurationNotification) instruction;
			contactReplica = REFITOrderGroups.getGroupPrimary(notification.groupID, notification.viewID);
			break;
		case EXECUTE_REQUEST:
			handleRequest(instruction);
			break;
		default:
			REFITLogger.logError(this, "drop instruction of unexpected type " + instruction.instructionType);
		}
	}

	@Override
	protected void stageComplete() {
		// check whether some requests can be processed
		processStableRequests();

		if (executionReplicas[replica.id] && agreementProgress) {
			helper.trackProgress(highestReceivedSeqNr, nextAgreementSeqNr, groupID);
		}
	}

	// ####################
	// # ORDERED REQUESTS #
	// ####################

	private final Map<Long, REFITOrderedCertificate> orderCertificates;
	private long nextAgreementSeqNr;
	private long highestReceivedSeqNr;
	private boolean agreementProgress;

	private void handleCheckpoint(REFITCheckpoint checkpoint) {
		updateViaCheckpoint(checkpoint.uid.seqNr);
		collectGarbage();
	}

	private void updateViaCheckpoint(long agreementSeqNr) {
		// only move forward as fast as the agreement replicas to reduce chance of replicas falling behind
		agreementSeqNr = Math.max(0, agreementSeqNr + REFITConfig.INSTANCES_AFTER_CHECKPOINT - REFITConfig.MAXIMUM_ORDER_INSTANCES_IN_PROGRESS);

		if (agreementSeqNr <= nextAgreementSeqNr) return;

		nextAgreementSeqNr = REFITOrderGroups.nextSequenceNumber(agreementSeqNr, groupID);
		highestReceivedSeqNr = agreementSeqNr;
		agreementProgress = true;
	}

	private void collectGarbage() {
		// Garbage collect requests
		orderCertificates.keySet().removeIf(aLong -> aLong < nextAgreementSeqNr);
	}

	private void handleRequest(REFITInstruction instruction) {
		REFITExecuteRequest operation = (REFITExecuteRequest) instruction;
		// only forward to execution replicas that have no collocated group member
		boolean[] usedExecutionReplicas = (REFITConfig.OPTIMIZED_AGREEMENT_AND_EXECUTION) ? nonGroupExecutionReplicas : executionReplicas;
		short replicaCount = (REFITConfig.OPTIMIZED_AGREEMENT_AND_EXECUTION) ? nonGroupReplicaCount : REFITConfig.EXEC_GROUP_SIZE;
		if (usedExecutionReplicas != null) {
			// create ordered request, send full request from primary, hash from others
			REFITOrderedRequest request = new REFITOrderedRequest(replica.id, operation,
					REFITConfig.USE_HASHED_ORDERED_REQUEST && (contactReplica == replica.id));

			request.setRecipientCount(replicaCount);
			request.serializeMessage();
			replica.mac.appendMulticastMAC(request, usedExecutionReplicas);
			replica.replicaMulticast(request, usedExecutionReplicas);

			if (REFITLogger.LOG_COUPLING)
				REFITLogger.logCoupling(this, "distribute ordered request " + request + " from operation " + operation);
		}

		// forward full local request
		REFITOrderedRequest fullRequest = new REFITOrderedRequest(replica.id, operation, false);
		fullRequest.markVerified();
		handleOrderedRequest(fullRequest);
	}

	private void handleOrderedRequest(REFITOrderedRequest request) {
		// non-execution replicas only accept own requests
		if (!executionReplicas[replica.id] && !request.isInternal()) return;
		if (request.uid.seqNr < nextAgreementSeqNr) return;

		// Get ordered certificate, override with local request
		boolean overrideCertificate = REFITConfig.OPTIMIZED_AGREEMENT_AND_EXECUTION && request.isInternal();
		REFITOrderedCertificate certificate = ensureCertificate(request.uid.seqNr, overrideCertificate);

		if (REFITLogger.LOG_COUPLING) REFITLogger.logCoupling(this, "received ordered request " + request);
		// only process further execute messages if the certificate is not yet stable
		if (!certificate.isStable()) {
			REFITOrderedCertificate.AddResult addResult = certificate.add(request, replica.mac);
			if (addResult.equals(REFITOrderedCertificate.AddResult.INVALID)) {
				REFITLogger.logWarning(this, "bad ORDERED_REQUEST (" + request + ")");
			}
		} else if (request.isExecutable() && !certificate.getStableRequest().isExecutable()) {
			if (!certificate.getStableRequest().setFullMessage(request)) {
				REFITLogger.logWarning(this, "non matching ORDERED_REQUEST (" + request + "), expected (" + certificate.getStableRequest() + ")");
			}
		}
		if (certificate.isStable()) {
			if (request.uid.seqNr > highestReceivedSeqNr) {
				highestReceivedSeqNr = request.uid.seqNr;
				agreementProgress = true;
			}
		}
	}

	private REFITOrderedCertificate ensureCertificate(long seqNr, boolean overrideCertificate) {
		REFITOrderedCertificate certificate = orderCertificates.get(seqNr);
		if (certificate == null || overrideCertificate) {
			if (REFITConfig.ENABLE_DEBUG_CHECKS && REFITOrderGroups.getGroupID(seqNr) != groupID) {
				throw new InternalError("Got request for wrong group " + seqNr + " expected " + groupID);
			}
			certificate = new REFITOrderedCertificate(seqNr, overrideCertificate || !executionReplicas[replica.id],
					REFITOrderGroups.getGroupByID(groupID));
			orderCertificates.put(seqNr, certificate);
		}
		return certificate;
	}

	private void processStableRequests() {
		while (true) {
			REFITOrderedCertificate certificate = orderCertificates.get(nextAgreementSeqNr);
			if (certificate == null || !certificate.isStable()) break;
			REFITOrderedRequest request = certificate.getStableRequest();
			if (request.isExecutable()) {
				// create and forward operation
				REFITExecuteRequest operation = new REFITExecuteRequest(request.uid.seqNr, certificate.getView(), request.message, null);
				requestSender.broadcast(operation);
			} else {
//              System.err.println("TODO: implement request fetching");
				if (!executionReplicas[replica.id]) {
					throw new InternalError("This should never happen " + request.toString());
				}
				break;
			}

			// move to next request
			orderCertificates.remove(nextAgreementSeqNr);
			// skip sequence numbers that don't belong to my group
			nextAgreementSeqNr = REFITOrderGroups.nextSequenceNumber(nextAgreementSeqNr, groupID);
		}
	}
}
