package refit.replica.stewardorder;

import java.math.BigInteger;
import java.util.*;

import refit.config.REFITConfig;
import refit.config.REFITLogger;
import refit.message.REFITInstruction;
import refit.message.REFITInstruction.REFITConfigurationNotification;
import refit.message.REFITInstruction.REFITExecuteRequest;
import refit.message.REFITMessage;
import refit.message.REFITOrderRequest;
import refit.replica.REFITBaseReplica;
import refit.replica.REFITObserver;
import refit.replica.REFITStage;
import refit.replica.order.REFITOrderProtocol;
import refit.replica.order.REFITOrderProtocolInstance;
import refit.replica.order.REFITOrderProtocolTransitionInstance;
import refit.replica.order.REFITOrderStageSlot;
import refit.replica.stewardorder.REFITStewardOrderMessages.REFITStewardAccept;
import refit.replica.stewardorder.REFITStewardOrderMessages.REFITStewardAcceptShare;
import refit.replica.stewardorder.REFITStewardOrderMessages.REFITStewardProposal;
import refit.scheduler.REFITSchedulerTaskType;
import threshsig.SigShare;

public class REFITStewardOrderStage extends REFITStage {

	public final REFITObserver<REFITExecuteRequest> requestSender;
	public final REFITObserver<REFITConfigurationNotification> configSender;
	private final int leaderSiteRepresentative;
	private final int leaderGroupID;
	private final int siteCount;

	private boolean[] localGroup;
	private boolean[] representatives;
	private boolean inLeaderGroup;

	private Slot[] slots;
	private long windowStart;

	private HashMap<Long, List<REFITMessage>> delayedMessages = new HashMap<>();

	public REFITStewardOrderStage(REFITBaseReplica replica, int leaderSiteRepresentative, int leaderGroupID, int siteCount) {
		super(REFITSchedulerTaskType.ORDER_STAGE, replica);
		this.leaderSiteRepresentative = leaderSiteRepresentative;
		this.leaderGroupID = leaderGroupID;
		this.inLeaderGroup = (REFITStewardGroups.stewardOrderGroupID(replica.id) == leaderGroupID);
		this.siteCount = siteCount;
		localGroup = REFITStewardGroups.stewardOrderGroup(replica.id);
		representatives = REFITStewardGroups.stewardRepresentatives();

		requestSender = new REFITObserver<>();
		configSender = new REFITObserver<>();
		slots = new Slot[REFITConfig.MAXIMUM_ORDER_INSTANCES_IN_PROGRESS];
		for (int i = 0; i < slots.length; i++) {
			slots[i] = new Slot();
		}

		windowStart = -REFITConfig.MAXIMUM_ORDER_INSTANCES_IN_PROGRESS;
		moveStartTo(0);
	}

	@Override
	public String toString() {
		return "STWORD";
	}

	@Override
	public void init() {
		// send configuration notification with stub protocol to execution
		configSender.broadcast(new REFITConfigurationNotification(
				new REFITOrderProtocol() {
					@Override
					public REFITOrderProtocolInstance createInstance(REFITOrderStageSlot slot) {
						throw new UnsupportedOperationException("Nope");
					}

					@Override
					public REFITOrderProtocolTransitionInstance createTransitionInstance(REFITOrderStageSlot slot) {
						throw new UnsupportedOperationException("Nope");
					}

					@Override
					public boolean useLightweightCheckpoints() {
						return false;
					}

					@Override
					public boolean hasCheckpointRecipients() {
						return true;
					}

					@Override
					public boolean hasUpdateRecipients() {
						return false;
					}
				}, -1, -1)
		);
	}

	@Override
	protected void handleMessage(REFITMessage message) {
		switch (message.groupType()) {
		case STEWARD_WAN:
			handleWAN(message.uid.seqNr, message);
			break;
		case INSTRUCTION:
			handleInstruction((REFITInstruction) message);
			break;
		default:
			REFITLogger.logError(this, "drop message of unexpected type " + message.type);
		}
	}

	private boolean messageInWindow(long seqNr) {
		return seqNr >= windowStart && seqNr < windowStart + slots.length;
	}

	private void storedFutureMessage(long seqNr, REFITMessage message) {
		if (seqNr < windowStart) return;
		List<REFITMessage> list = delayedMessages.computeIfAbsent(seqNr, k -> new ArrayList<>());
		list.add(message);
	}

	private Slot getSlot(long seqNr) {
		return slots[(int) (seqNr % slots.length)];
	}

	private void handleWAN(long seqNr, REFITMessage message) {
		if (!messageInWindow(seqNr)) {
			storedFutureMessage(seqNr, message);
			return;
		}

		getSlot(seqNr).handleMessage(message);
	}

	private void handleInstruction(REFITInstruction message) {
		switch (message.instructionType) {
		case EXECUTE_REQUEST:
			handleWAN(((REFITExecuteRequest) message).agreementSeqNr, message);
			break;
		case PROGRESS_NOTIFICATION:
			REFITInstruction.REFITProgressNotification progress = (REFITInstruction.REFITProgressNotification) message;
			moveStartTo(progress.agreementSeqNr);
			break;
		default:
			REFITLogger.logError(this, "drop instruction of unexpected type " + message.instructionType);
		}
	}

	private void moveStartTo(long newSeqNr) {
		long newStart = Math.max(windowStart, newSeqNr);
		long newEnd = newStart + slots.length;
		long oldEnd = windowStart + slots.length;
		long initFrom = Math.max(oldEnd, newStart);
		windowStart = newStart;

		for (long seqNr = initFrom; seqNr < newEnd; seqNr++) {
			Slot slot = slots[(int) (seqNr % slots.length)];
			slot.init(seqNr);
			if (delayedMessages.containsKey(seqNr)) {
				for (REFITMessage message : delayedMessages.get(seqNr)) {
					slot.handleMessage(message);
				}
			}
		}

		delayedMessages.keySet().removeIf(k -> k < windowStart);
	}

	enum StewardStep {
		PRE_PROPOSAL,
		PRE_ACCEPT,
		COMPLETE
	}

	private class Slot {
		long seqNr;
		REFITStewardProposalCertificate propCert;
		REFITStewardAcceptCertificate acceptCert;
		StewardStep state;
		Queue<REFITStewardAccept> acceptQueue = new ArrayDeque<>();

		public Slot() {
			int leaderGroupID = REFITStewardGroups.stewardOrderGroupID(leaderSiteRepresentative);
			propCert = new REFITStewardProposalCertificate(leaderGroupID,
					REFITStewardGroups.stewardOrderGroupID(replica.id), REFITConfig.ORDER_GROUP_SIZE);
			acceptCert = new REFITStewardAcceptCertificate(leaderGroupID, siteCount);
		}

		public void init(long seqNr) {
			this.seqNr = seqNr;
			propCert.init(seqNr, 0);
			acceptCert.init(seqNr, 0);
			state = StewardStep.PRE_PROPOSAL;
			acceptQueue.clear();
		}

		public void handleMessage(REFITMessage message) {
			switch (message.type) {
			case STEWARD_PROPOSAL:
				handleProposal((REFITStewardProposal) message);
				break;
			case STEWARD_ACCEPT_SHARE:
				handleAcceptShare((REFITStewardAcceptShare) message);
				break;
			case STEWARD_ACCEPT:
				handleAccept((REFITStewardAccept) message);
				break;
			case INSTRUCTION:
				handleInstruction((REFITInstruction) message);
				break;
			default:
				REFITLogger.logError(this, "drop message of unexpected type " + message.type);
			}

			checkStateTransition();
		}

		private boolean isGroupRepresentative(int replicaId) {
			return replicaId >= 0 && replicaId < REFITConfig.TOTAL_NR_OF_REPLICAS && (replicaId % REFITConfig.ORDER_GROUP_SIZE) == 0;
		}

		private void handleProposal(REFITStewardProposal message) {
			if (!propCert.addPrePrepare(message, replica.mac)) {
				REFITLogger.logWarning(this, "Ignoring invalid proposal " + message);
				return;
			}
			if (!acceptCert.addPrePrepare(message, replica.mac)) {
				REFITLogger.logWarning(this, "Ignoring invalid proposal2 " + message);
				return;
			}

			if (isGroupRepresentative(replica.id) && replica.id != leaderSiteRepresentative) {
				replica.replicaMulticast(message, localGroup);
			}

			if (REFITStewardGroups.stewardOrderGroupID(replica.id) == leaderGroupID) {
				// the leader group does not create accepts
				// A-11 C1. is only triggered when a proposal is received from the network which does not happen
				// for the leader site
				return;
			}

			SigShare share = replica.mac.calcSigShare(propCert.prePrepare.calcShortHash());

			// Create and distribute COMMIT
			REFITStewardAcceptShare ownCommit = new REFITStewardAcceptShare(propCert.prePrepare, replica.id, share);
			ownCommit.serializeMessage();
			replica.mac.appendSignature(ownCommit);
			replica.replicaMulticast(ownCommit, localGroup);

			handleAcceptShare(ownCommit);
		}

		private void handleAcceptShare(REFITStewardAcceptShare message) {
			if (propCert.isCommitted()) {
				// already committed
				return;
			}

			// Add COMMIT to certificate
			REFITStewardProposalCertificate.AddResult addResult = propCert.addCommit(message, replica.mac);
			if (addResult.equals(REFITStewardProposalCertificate.AddResult.INVALID)) {
				REFITLogger.logWarning(this, "bad COMMIT " + message);
				return;
			}
		}

		private void handleAccept(REFITStewardAccept message) {
			if (propCert.prePrepare == null) {
				acceptQueue.add(message);
				return;
			}
			if (acceptCert.isCommitted()) {
				// request already completed
				return;
			}

			if (acceptCert.addCommit(message, replica.mac).equals(REFITStewardAcceptCertificate.AddResult.INVALID)) {
				REFITLogger.logWarning(this, "bad ACCEPT " + message);
				return;
			}

			if (isGroupRepresentative(replica.id)) {
				replica.replicaMulticast(message, localGroup);
			}
		}

		private void checkStateTransition() {
			if (state == StewardStep.PRE_PROPOSAL) {
				if (propCert.isCommitted()) {
					REFITStewardAccept accept = new REFITStewardAccept(propCert.prePrepare, replica.id, propCert.getCombinedSig());
					accept.serializeMessage();
					replica.mac.appendNullMAC(accept);
					if (isGroupRepresentative(replica.id)) {
						replica.replicaMulticast(accept, representatives);
					}
					handleAccept(accept);
					state = StewardStep.PRE_ACCEPT;
				} else if (inLeaderGroup && propCert.prePrepare != null) {
					state = StewardStep.PRE_ACCEPT;
				}
			}
			if (state == StewardStep.PRE_ACCEPT) {
				while (!acceptQueue.isEmpty()) {
					REFITStewardAccept accept = acceptQueue.poll();
					handleAccept(accept);
				}
				if (acceptCert.isCommitted()) {
					requestSender.broadcast(new REFITExecuteRequest(acceptCert.prePrepare.uid.seqNr, 0,
							acceptCert.prePrepare.message, null));
					state = StewardStep.COMPLETE;
				}
			}
		}

		private void handleInstruction(REFITInstruction message) {
			switch (message.instructionType) {
			case EXECUTE_REQUEST:
				startWithProposal((REFITExecuteRequest) message);
				break;
			default:
				REFITLogger.logError(this, "drop instruction of unexpected type " + message.instructionType);
			}
		}

		private void startWithProposal(REFITExecuteRequest message) {
			if (REFITStewardGroups.stewardOrderGroupID(replica.id) != leaderGroupID) {
				return;
			}

			BigInteger combinedSig = (BigInteger) message.attachment;

			// According to A-7 A5. the proposal is not explicitly broadcast in the leader site
			// According to A-11 A6. only the leader site repr. calls ASSIGN-SEQUENCE
			REFITStewardProposal proposal = new REFITStewardProposal(message.agreementSeqNr, 0,
					(short) leaderSiteRepresentative, (REFITOrderRequest) message.message, combinedSig);
			proposal.serializeMessage();
			// broadcast to other site
			replica.mac.appendNullMAC(proposal);

			if (replica.id == leaderSiteRepresentative) {
				replica.replicaMulticast(proposal, representatives);
			}

			handleMessage(proposal);
		}
	}
}
